@@ -4,30 +4,28 @@
  * Project: GGZ Core Client Lib
  *          Modified from confio for use by server (rgade - 08/06/01)
  * Date: 11/27/00
- * $Id: conf.c,v 1.18 2003/02/04 00:53:21 jdorje Exp $
+ * $Id: conf.c 7885 2006-03-06 18:46:12Z josef $
  *
  * Internal functions for handling configuration files
  *
  * Copyright (C) 2000-2002 Brent Hendricks.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#ifdef HAVE_CONFIG_H
-#  include <config.h>
-#endif
+#include "config.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -58,6 +56,10 @@
  *	b) the value (variable data)
  */
 
+#ifdef MKDIR_TAKES_ONE_ARG
+#define mkdir(a, b) (mkdir) (a)
+#endif
+
 
 /****
  **** Structure definitions for our two list element types
@@ -83,10 +85,10 @@ typedef struct conf_entry_t {
 /* Our private functions and vars */
 static GGZList * file_parser(const char *path);
 static void parse_line(char *line, char **varname, char **varvalue);
-static int section_compare(void *a, void *b);
+static int section_compare(const void *a, const void *b);
 static void *section_create(void *data);
 static void section_destroy(void *data);
-static int entry_compare(void *a, void *b);
+static int entry_compare(const void *a, const void *b);
 static void *entry_create(void *data);
 static void entry_destroy(void *data);
 static conf_file_t * get_file_data(int handle);
@@ -268,7 +270,8 @@ int ggz_conf_write_string(int handle, const char *section,
 
 	/* Is this confio writeable? */
 	if(!f_data->writeable) {
-		_ggz_debug("CONF", "ggz_conf_write_string: file is read-only");
+		ggz_debug(GGZ_CONF_DEBUG,
+			  "ggz_conf_write_string: file is read-only");
 		return -1;
 	}
 
@@ -277,7 +280,7 @@ int ggz_conf_write_string(int handle, const char *section,
 	if(s_entry == NULL) {
 		/* We need to create a new [Section] */
 		if(ggz_list_insert(f_data->section_list, (void*)section) < 0) {
-			_ggz_debug("CONF",
+			ggz_debug(GGZ_CONF_DEBUG,
 				  "ggz_conf_write_string: insertion error");
 			return -1;
 		}
@@ -289,7 +292,8 @@ int ggz_conf_write_string(int handle, const char *section,
 	e_data.key = (char*)key;
 	e_data.value = (char*)value;
 	if(ggz_list_insert(s_data->entry_list, &e_data) < 0) {
-		_ggz_debug("CONF", "ggz_conf_write_string: insertion error");
+		ggz_debug(GGZ_CONF_DEBUG,
+			  "ggz_conf_write_string: insertion error");
 			  
 		return -1;
 	}
@@ -381,7 +385,8 @@ int ggz_conf_remove_section(int handle, const char *section)
 
 	/* Is this confio writeable? */
 	if(!f_data->writeable) {
-		_ggz_debug("CONF", "ggz_conf_remove_section: file is read-only");
+		ggz_debug(GGZ_CONF_DEBUG,
+			  "ggz_conf_remove_section: file is read-only");
 			  
 		return -1;
 	}
@@ -421,8 +426,8 @@ int ggz_conf_remove_key(int handle, const char *section, const char *key)
 
 	/* Is this confio writeable? */
 	if(!f_data->writeable) {
-		_ggz_debug("CONF",
-		      "ggzcore_confio_remove_key: file is read-only");
+		ggz_debug(GGZ_CONF_DEBUG,
+			  "ggzcore_confio_remove_key: file is read-only");
 		return -1;
 	}
 
@@ -469,15 +474,15 @@ int ggz_conf_commit(int handle)
 
 	/* Is this confio writeable? */
 	if(!f_data->writeable) {
-		_ggz_debug("CONF",
-		      "ggzcore_confio_commit: file is read-only");
+		ggz_debug(GGZ_CONF_DEBUG,
+			  "ggzcore_confio_commit: file is read-only");
 		return -1;
 	}
 
 	/* Open our configuration file for writing */
 	if((c_file = fopen(f_data->path, "w")) == NULL) {
-		_ggz_debug("CONF", "Unable to write config file %s",
-			      f_data->path);
+		ggz_debug(GGZ_CONF_DEBUG, "Unable to write config file %s",
+			  f_data->path);
 		return -1;
 	}
 
@@ -693,7 +698,7 @@ static GGZList * file_parser(const char *path)
 	GGZListEntry	*s_entry;
 	conf_section_t	*s_data=NULL;
 	conf_entry_t	*e_data=NULL;
-
+	struct stat		st;
 
 	/* Create the section list */
 	s_list = ggz_list_create(section_compare,
@@ -703,6 +708,12 @@ static GGZList * file_parser(const char *path)
 	if(!s_list)
 		return NULL;
 
+	/* Check for file mode */
+	if((stat(path, &st)) || (!S_ISREG(st.st_mode))) {
+		ggz_error_msg("File %s is not a regular file", path);
+		return NULL;
+	}
+
 	/* Open the input config file */
 	if((c_file = open(path, O_RDONLY)) == -1) {
 		/* This should be impossible now, due to checks made earlier? */
@@ -715,6 +726,7 @@ static GGZList * file_parser(const char *path)
 
 	/* Setup some temp storage to use */
 	e_data = ggz_malloc(sizeof(conf_entry_t));
+	varvalue = NULL;
 
 	/* Read individual lines and pass them off to be parsed */
 	while((line = ggz_read_line(c_struct)) != NULL) {
@@ -856,8 +868,8 @@ static conf_file_t * get_file_data(int handle)
 		f_entry = ggz_list_next(f_entry);
 	}
 	if(f_entry == NULL) {
-		_ggz_debug("CONF",
-			      "get_file_data:  Invalid conf handle requested");
+		ggz_debug(GGZ_CONF_DEBUG,
+			  "get_file_data:  Invalid conf handle requested");
 		f_data = NULL;
 	}
 
@@ -867,20 +879,26 @@ static conf_file_t * get_file_data(int handle)
 
 /* The following three functions perform list comparisons  */
 /* and data creation and destruction for the section lists */
-static int section_compare(void *a, void *b)
+static int section_compare(const void *a, const void *b)
 {
-	/* Note that this function is a little odd since the 'b' passed */
-	/* is not expected to be the full struct, but just the name str */
-	conf_section_t	*s_a;
+	/* Note that this function is a little odd since the 'b' passed
+	   is not expected to be the full struct, but just the name str.
+	   This is because the whole section list is a hack and every
+	   time an insertion or lookup is done all that's passed in is the
+	   name string.  The callback functions here are expected to do
+	   the rest of the work of making the section.  See also
+	   section_destroy. */
+	const conf_section_t *s_a = a;
+	const char *s_b = b;
 
-	s_a = a;
-	return strcmp(s_a->name, b);
+	return strcmp(s_a->name, s_b);
 }
 
 static void *section_create(void *data)
 {
-	/* Note that this function is a little odd since the data passed */
-	/* is not expected to be the full struct, but just the name str  */
+	/* Note that this function is a little odd since the data passed
+	   is not expected to be the full struct, but just the name str.
+	   See the comment in section_compare. */
 	conf_section_t	*dst;
 
 	dst = ggz_malloc(sizeof(conf_section_t));
@@ -913,12 +931,9 @@ static void section_destroy(void *data)
 
 /* The following three functions perform list comparisons */
 /* and data creation and destruction for the entry lists  */
-static int entry_compare(void *a, void *b)
+static int entry_compare(const void *a, const void *b)
 {
-	conf_entry_t	*e_a, *e_b;
-
-	e_a = a;
-	e_b = b;
+	const conf_entry_t *e_a = a, *e_b = b;
 
 	return strcmp(e_a->key, e_b->key);
 }
@@ -952,31 +967,36 @@ static void entry_destroy(void *data)
  */
 int make_path(const char *full, mode_t mode)
 {
-	char		*copy, *node, *path;
-	struct stat	stats;
+	struct stat stats;
+	size_t len = strlen(full) + 1;
+	char copy_buf[len], path[len];
+	char *copy = copy_buf;
 
-	copy = ggz_strdup(full);
+	strcpy(copy, full);
+	path[0] = '\0';
 
-	/* Allocate and zero memory for path */
-	path = ggz_malloc(strlen(full)+1);
- 
 	/* Skip preceding / */
 	if (copy[0] == '/')
 		copy++;
 
-	while ((node = strsep(&copy, "/"))) {
+	do {
+		char *next = strchr(copy, '/');
+
+		/* If this is the last token, it's the file name - break */
+		if (!next) break;
+		
+		*next = '\0';
+
 		/* While there's still stuff left, it's a directory */
-		if (copy != NULL) {
-			strcat(strcat(path, "/"), node);
-			if (mkdir(path, mode) < 0
-			    && (stat(path, &stats) < 0 || !S_ISDIR(stats.st_mode))) {
-				ggz_free(path);
-				ggz_free(copy);
-				
-				return -1;
-			}
+		strcat(strcat(path, "/"), copy);
+
+		if (mkdir(path, mode) < 0
+		    && (stat(path, &stats) < 0 || !S_ISDIR(stats.st_mode))) {
+			return -1;
 		}
-	}
+
+		copy = next + 1;
+	} while (1);
 
 	return 0;
 }
