@@ -3,7 +3,7 @@
  * Author: Brent Hendricks
  * Project: libeasysock
  * Date: 4/16/98
- * $Id: easysock.c,v 1.13 2003/05/10 14:32:19 dr_maux Exp $
+ * $Id: easysock.c 7877 2006-02-27 12:15:43Z josef $
  *
  * A library of useful routines to make life easier while using 
  * sockets
@@ -22,25 +22,47 @@
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#ifdef HAVE_CONFIG_H
-#  include <config.h>
+/*
+ * actually used definitions:
+HAVE_WINSOCK2_H
+GGZ_HAVE_PF_LOCAL
+SUN_LEN (local substitute)
+  -> also in support.h!, therefore deleted
+GGZ_HAVE_SENDMSG
+HAVE_MSGHDR_MSG_CONTROL
+  HAVE_CMSG_* -> support.h
+  CMSG_* -> support.h
+DEBUG_MEM
+*/
+
+#include "config.h"
+
+#ifdef GAI_A
+#define _GNU_SOURCE
 #endif
 
 #ifdef DEBUG_MEM
-# include <dmalloc.h>
+#include <dmalloc.h>
 #endif
 
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/un.h>
-#include <sys/uio.h>
 #include <sys/param.h>
 
+#ifdef HAVE_WINSOCK2_H
+#include <winsock2.h>
+#include <ws2tcpip.h>
+#else
+#include <sys/socket.h>
+#include <sys/uio.h>
+#include <sys/un.h>
+#include <arpa/inet.h>
 #include <netinet/in.h>
 #include <netdb.h>
+#endif
+
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
@@ -53,10 +75,14 @@
 #include "support.h"
 
 #define SA struct sockaddr  
+#define HOSTNAMELEN 1024
 
 static ggzIOError _err_func = NULL;
 static ggzIOExit _exit_func = exit;
+static ggzNetworkNotify _notify_func = NULL;
 static unsigned int ggz_alloc_limit = 32768;
+static unsigned int ggz_socketcreation = 0;
+static unsigned int ggz_socketport = 0;
 
 
 int ggz_set_io_error_func(ggzIOError func)
@@ -103,30 +129,207 @@ unsigned int ggz_set_io_alloc_limit(const unsigned int limit)
 }
 
 
-int ggz_make_socket(const GGZSockType type, const unsigned short port, 
-		   const char *server)
+static void ggz_network_shutdown(void)
 {
-	int sock;
-	const int on = 1;
-	struct sockaddr_in name;
-	struct hostent *hp;
+#ifdef HAVE_WINSOCK2_H
+	WSACleanup();
+#endif
+}
 
-	if ( (sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
-		if (_err_func)
-			(*_err_func) (strerror(errno), GGZ_IO_CREATE, -1, GGZ_DATA_NONE);
+
+/* Do network initialization. */
+int ggz_init_network(void)
+{
+	static int initialized = 0;
+
+	if (!initialized) {
+
+#ifdef HAVE_WINSOCK2_H
+		WSADATA wsa;
+		if (WSAStartup(MAKEWORD(1, 1), &wsa) !=0 ){
+			return -1;
+		}
+#endif
+
+		initialized = 1;
+
+		atexit(ggz_network_shutdown);
+	}
+
+	return 0;
+}
+
+
+static int es_bind(const char *host, int port)
+{
+	int sockfd;
+	const int on = 1;
+	/* FIXME: better test on getaddrinfo() directly */
+#ifdef HAVE_WINSOCK2_H
+	struct sockaddr_in name;
+
+	sockfd = socket(PF_INET, SOCK_STREAM, 0);
+
+	if (sockfd < 0) {
+		if (_err_func) {
+			const char *msg = "could-not-create-socket-error";
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
 		return -1;
 	}
 
+	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+
+	memset(&name, 0, sizeof(name));
 	name.sin_family = AF_INET;
-	name.sin_port = htons(port);
+	name.sin_port = port;
+	name.sin_addr.s_addr = htonl(INADDR_ANY);
+
+	if (bind(sockfd, &name, sizeof(name)) != 0) {
+		if (_err_func) {
+			const char *msg = strerror(errno);
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
+		close(sockfd);
+		sockfd = -1;
+	}
+#else
+	int n;
+	struct addrinfo hints, *res, *ressave;
+	char serv[30];
+
+	snprintf(serv, sizeof(serv), "%d", (unsigned int)port);
+
+	memset(&hints, 0, sizeof(struct addrinfo));
+	hints.ai_flags = AI_PASSIVE;
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_STREAM;
+
+	if ((n = getaddrinfo(host, serv, &hints, &res)) != 0) {
+		if (_err_func) {
+			const char *msg = gai_strerror(n);
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
+		return -1;
+	}
+
+	ressave = res;
+
+	do {
+		sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+		if (sockfd < 0)
+			continue;
+
+		setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
+
+		if (bind(sockfd, res->ai_addr, res->ai_addrlen) == 0)
+			break;
+
+		close(sockfd);
+		sockfd = -1;
+
+	} while ( (res = res->ai_next) != NULL);
+
+	freeaddrinfo(ressave);
+#endif
+
+	return(sockfd);
+}
+
+
+static int es_connect(const char *host, int port)
+{
+	int sockfd;
+#ifdef HAVE_WINSOCK2_H
+	struct hostent *h;
+	struct sockaddr_in name;
+
+	sockfd = socket(PF_INET, SOCK_STREAM, 0);
+
+	if (sockfd < 0) {
+		if (_err_func) {
+			const char *msg = "could-not-create-socket-error";
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
+		return -1;
+	}
+
+	h = gethostbyname(host);
+	if (!h) {
+		if (_err_func) {
+			const char *msg = "could-not-create-socket-error";
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
+		return -1;
+	}
+
+	memset(&name, 0, sizeof(name));
+	name.sin_family = AF_INET;
+	name.sin_port = port;
+	memcpy(&name.sin_addr, h->h_addr, h->h_length);
+
+	if (connect(sockfd, &name, sizeof(name)) != 0) {
+		if (_err_func) {
+			const char *msg = strerror(errno);
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
+		close(sockfd);
+		sockfd = -1;
+	}
+#else
+	int n;
+	struct addrinfo hints, *res, *ressave;
+	char serv[30];
+
+	snprintf(serv, sizeof(serv), "%d", (unsigned int)port);
+
+	memset(&hints, 0, sizeof(struct addrinfo));
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_STREAM;
+
+	if ((n = getaddrinfo(host, serv, &hints, &res)) != 0) {
+		if (_err_func) {
+			const char *msg = gai_strerror(n);
+
+			(*_err_func) (msg, GGZ_IO_CREATE, 0, GGZ_DATA_NONE);
+		}
+		return -1;
+	}
+
+	ressave = res;
+
+	do {
+		sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+		if (sockfd < 0)
+			continue;
+		if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0)
+			break;
+
+		close(sockfd);
+		sockfd = -1;
+
+	} while ( (res = res->ai_next) != NULL);
+
+	freeaddrinfo(ressave);
+#endif
+
+	return(sockfd);
+}
+
+
+int ggz_make_socket(const GGZSockType type, const unsigned short port, 
+		   const char *server)
+{
+	int sock = -1;
+
+	if (ggz_init_network() < 0) { /* Just in case. */
+		return -1;
+	}
 
 	switch (type) {
 
 	case GGZ_SOCK_SERVER:
-		name.sin_addr.s_addr = htonl(INADDR_ANY);
-		if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void*)&on, 
-			       sizeof(on)) < 0
-		    || bind(sock, (SA *)&name, sizeof(name)) < 0) {
+		if ( (sock = es_bind(server, port)) < 0) {
 			if (_err_func)
 				(*_err_func) (strerror(errno), GGZ_IO_CREATE, 
 					      sock, GGZ_DATA_NONE);
@@ -135,15 +338,13 @@ int ggz_make_socket(const GGZSockType type, const unsigned short port,
 		break;
 
 	case GGZ_SOCK_CLIENT:
-		if ( (hp = gethostbyname(server)) == NULL) {
-			if (_err_func)
-				(*_err_func) ("Lookup failure", GGZ_IO_CREATE, 
-					      sock, GGZ_DATA_NONE);
-			return -2;
-			break;
+		if ((_notify_func) && (!ggz_socketcreation)) {
+			ggz_socketcreation = 1;
+			ggz_socketport = port;
+			ggz_resolvename(server);
+			return -3;
 		}
-		memcpy(&name.sin_addr, hp->h_addr, hp->h_length);
-		if (connect(sock, (SA *)&name, sizeof(name)) < 0) {
+		if ( (sock = es_connect(server, port)) < 0) {
 			if (_err_func)
 				(*_err_func) (strerror(errno), GGZ_IO_CREATE, 
 					      sock, GGZ_DATA_NONE);
@@ -167,11 +368,13 @@ int ggz_make_socket_or_die(const GGZSockType type, const unsigned short port,
 	return sock;
 }
 
-
-int ggz_make_unix_socket(const GGZSockType type, const char* name) 
+int ggz_make_unix_socket(const GGZSockType type, const char* name)
 {
+#if GGZ_HAVE_PF_LOCAL
 	int sock;
 	struct sockaddr_un addr;
+
+	ggz_init_network(); /* Just in case. */
 	
 	if ( (sock = socket(PF_LOCAL, SOCK_STREAM, 0)) < 0) {
 		if (_err_func)
@@ -206,17 +409,24 @@ int ggz_make_unix_socket(const GGZSockType type, const char* name)
 		break;
 	}
 	return sock;
+#else
+	return -1;
+#endif
 }
 
 
 int ggz_make_unix_socket_or_die(const GGZSockType type, const char* name) 
 {
+#if GGZ_HAVE_PF_LOCAL
 	int sock;
 	
 	if ( (sock = ggz_make_unix_socket(type, name)) < 0)
 		(*_exit_func) (-1);
 	
 	return sock;
+#else
+	return -1;
+#endif
 }
 
 
@@ -396,7 +606,7 @@ int ggz_read_string(const int sock, char *message, const unsigned int len)
 	unsigned int size;
 	int status;
 
-	if (ggz_read_int(sock, &size) < 0)
+	if (ggz_read_int(sock, (int*)&size) < 0)
 		return -1;
 	
 	if (size > len) {
@@ -443,7 +653,7 @@ int ggz_read_string_alloc(const int sock, char **message)
 	unsigned int size;
 	int status;
 
-	if (ggz_read_int(sock, &size) < 0)
+	if (ggz_read_int(sock, (int*)&size) < 0)
 		return -1;
 
 	if (size > ggz_alloc_limit) {
@@ -498,7 +708,12 @@ int ggz_writen(const int sock, const void *vptr, size_t n)
 	ptr = vptr;
 	nleft = n;
 	while (nleft > 0) {
-		if ((nwritten = write(sock, ptr, nleft)) <= 0) {
+#ifdef HAVE_WINSOCK2_H
+		nwritten = send(sock, ptr, nleft, 0);
+#else
+		nwritten = write(sock, ptr, nleft);
+#endif
+		if (nwritten <= 0) {
 			if (errno == EINTR)
 				nwritten = 0;	/* and call write() again */
 			else
@@ -524,7 +739,12 @@ int ggz_readn(const int sock, void *vptr, size_t n)
 	ptr = vptr;
 	nleft = n;
 	while (nleft > 0) {
-		if ((nread = read(sock, ptr, nleft)) < 0) {
+#ifdef HAVE_WINSOCK2_H
+		nread = recv(sock, ptr, nleft, 0);
+#else
+		nread = read(sock, ptr, nleft);
+#endif
+		if (nread < 0) {
 			if (errno == EINTR)
 				nread = 0;	/* and call read() again */
 			else
@@ -542,7 +762,8 @@ int ggz_readn(const int sock, void *vptr, size_t n)
 
 int ggz_write_fd(int sock, int sendfd)
 {
-        struct msghdr msg;
+#if GGZ_HAVE_SENDMSG
+	struct msghdr msg;
 	struct iovec iov[1];
 
 #ifdef	HAVE_MSGHDR_MSG_CONTROL
@@ -552,6 +773,13 @@ int ggz_write_fd(int sock, int sendfd)
 	} control_un;
 	struct cmsghdr *cmptr;
 
+	/* Zero out the msg struct.  We should really initialize it properly.
+	   There are some extra fields that don't get initialized below, but
+	   I don't know if it's portable to add initializers for them.  Try
+	   removing this line and running under valgrind to see the error.
+	   See also the memset down below. */
+	memset(&msg, 0, sizeof(msg));
+
 	msg.msg_control = control_un.control;
 	msg.msg_controllen = sizeof(control_un.control);
 
@@ -561,6 +789,9 @@ int ggz_write_fd(int sock, int sendfd)
 	cmptr->cmsg_type = SCM_RIGHTS;
 	*((int *) CMSG_DATA(cmptr)) = sendfd;
 #else
+	/* See the comment for memset() above. */
+	memset(&msg, 0, sizeof(msg));
+
 	msg.msg_accrights = (caddr_t) &sendfd;
 	msg.msg_accrightslen = sizeof(int);
 #endif
@@ -584,11 +815,15 @@ int ggz_write_fd(int sock, int sendfd)
 
 	ggz_debug(GGZ_SOCKET_DEBUG, "Sent \"%d\" : fd.", sendfd);
 	return 0;
+#else
+	return -1;
+#endif
 }
 
 
 int ggz_read_fd(int sock, int *recvfd)
 {
+#if GGZ_HAVE_SENDMSG
 	struct msghdr msg;
 	struct iovec iov[1];
 	ssize_t	n;
@@ -675,5 +910,128 @@ int ggz_read_fd(int sock, int *recvfd)
 	
 	ggz_debug(GGZ_SOCKET_DEBUG, "Received \"%d\" : fd.", *recvfd);
         return 0;
+#else
+	return -1;
+#endif
+}
+
+int ggz_set_network_notify_func(ggzNetworkNotify func)
+{
+	_notify_func = func;
+	return 0;
+}
+
+#ifdef GAI_A
+static struct gaicb *global_req = NULL;
+
+static void ggz_resolved(sigval_t arg)
+{
+	char hostname[64];
+	struct addrinfo *res;
+	int ret;
+	struct gaicb *req;
+
+	req = global_req;
+	if(gai_error(req)) {
+		(*_notify_func)(req->ar_name, -2);
+		return;
+	}
+	res = req->ar_result;
+	ret = getnameinfo(res->ai_addr, res->ai_addrlen,
+		hostname, sizeof(hostname), NULL, 0, NI_NUMERICHOST);
+	if(ret) {
+		(*_notify_func)(req->ar_name, -1);
+		return;
+	}
+	if(!ggz_socketcreation) {
+		(*_notify_func)(hostname, -1);
+	} else {
+		ret = ggz_make_socket(GGZ_SOCK_CLIENT, ggz_socketport, hostname);
+		ggz_socketcreation = 0;
+		(*_notify_func)(hostname, ret);
+	}
+}
+#endif
+
+const char *ggz_resolvename(const char *name)
+{
+	if (_notify_func) {
+#ifdef GAI_A
+	  	/* Start asynchronous lookup */
+	  	struct sigevent sigev;
+		struct gaicb *req;
+		int ret;
+
+		req = ggz_malloc(sizeof(*req));
+		req->ar_name = name;
+		req->ar_service = NULL;
+		req->ar_request = NULL;
+		sigev.sigev_notify = SIGEV_THREAD;
+		sigev.sigev_value.sival_ptr = NULL;
+		sigev.sigev_notify_function = ggz_resolved;
+		sigev.sigev_notify_attributes = NULL;
+		global_req = req;
+		ret = getaddrinfo_a(GAI_NOWAIT, &req, 1, &sigev);
+		if (ret) {
+		  	(*_notify_func)(name, -2);
+		}
+#else
+		/* Pass back unresolved name */
+		(*_notify_func)(name, -2);
+#endif
+		return NULL;
+	} else {
+		return name;
+	}
+}
+
+const char *ggz_getpeername(int fd, int resolve)
+{
+	char *ip;
+
+#ifndef HAVE_WINSOCK2_H
+	struct sockaddr_storage addr;
+	socklen_t addrsize;
+	int ret;
+
+	addrsize = sizeof(addr);
+	ret = getpeername(fd, (struct sockaddr*)&addr, &addrsize);
+	if(ret != 0)
+	{
+		return NULL;
+	}
+
+	if(resolve)
+	{
+		ip = (char*)ggz_malloc(HOSTNAMELEN);
+		ret = getnameinfo((struct sockaddr*)&addr, addrsize,
+			ip, HOSTNAMELEN, NULL, 0, NI_NUMERICHOST);
+		if(ret != 0)
+		{
+				ggz_free(ip);
+				return NULL;
+		}
+		return ip;
+	}
+
+	/* FIXME: IPv4 compatibility? One could use getnameinfo() on addr... */
+	if(addr.ss_family == AF_INET6)
+	{
+		ip = (char*)ggz_malloc(INET6_ADDRSTRLEN);
+		inet_ntop(AF_INET6, (void*)&(((struct sockaddr_in6*)&addr)->sin6_addr),
+			ip, INET6_ADDRSTRLEN);
+	}
+	else if(addr.ss_family == AF_INET)
+	{
+		ip = (char*)ggz_malloc(INET_ADDRSTRLEN);
+		inet_ntop(AF_INET, (void*)&(((struct sockaddr_in*)&addr)->sin_addr),
+			ip, INET_ADDRSTRLEN);
+	}
+	else ip = NULL;
+#else
+	ip = NULL;
+#endif
+
+	return ip;
 }
 
